@@ -1,64 +1,101 @@
-import typing
-from threading import Lock, Thread, Condition
-
-import websocket
-import json
+from dataclasses import dataclass, asdict
 from functools import reduce
+import json
+from threading import Condition, Lock, Thread
+from typing import Callable, Dict, Generic, List, TypeVar, TypedDict, Union
+from urllib.parse import urlparse
+import websocket
 
-from dataclasses import dataclass
+from .config import Config
 
-from .interceptors.retry_response import RetryResponseInterceptor
+T = TypeVar("T")
+
+
+@dataclass
+class RequestObject:
+    # Identifier established by the initiator of the request
+    id: int
+    # name of the engine method.
+    method: str
+    # target of the method.
+    handle: int
+    # the parameters can be provided by name through an object or by position through an array
+    params: List[any]
+    # version of JSON-RPC defaults to 2.0
+    jsonrpc: str = "2.0"
+
+
+@dataclass
+class ResponseObject:
+    # id of the backend object.
+    id: int
+    #  QIX type of the backend object. Can for example be "Doc" or "GenericVariable".
+    type: str
+    # Custom type of the backend object, if defined in qInfo.
+    genericType: str
+    # Handle of the backend object.
+    handle: int
+    # represents the returned value from engine
+    result: dict
+
+
+RequestInterceptor = Callable[[RequestObject], RequestObject]
+""" RPC Request interceptor """
+ResponseInterceptor = Callable[[dict], dict]
+""" RPC Request interceptor """
+
+
+class InterceptorHandler(Generic[T]):
+    handlers: List[T]
+    "list containing the interceptors"
+
+    def __init__(self):
+        self.handlers = []
+
+    def use(self, interceptor: Union[T, List[T]]) -> None:
+        """
+        method helper for registering an interceptor
+
+        Parameters
+        ----------
+        interceptor: function interceptor for requests/responses
+        """
+        if isinstance(interceptor, list):
+            self.handlers = self.handlers + interceptor
+        else:
+            self.handlers.append(interceptor)
+
+
+class Interceptors(TypedDict):
+    response: InterceptorHandler[ResponseInterceptor]
+    request: InterceptorHandler[RequestInterceptor]
 
 
 class FailedToConnect(Exception):
     pass
 
 
-@dataclass
-class RequestData:
-    jsonrpc: str
-    id: int
-    method: str
-    handle: int
-    params: typing.List[any]
+class RpcSession:
 
+    _interceptors: Interceptors
 
-RequestInterceptor = typing.Callable[[RequestData], RequestData]
-ResponseInterceptor = typing.Callable[[dict], dict]
+    _ws_url: str
+    _headers: List[str]
 
-
-class RpcClient:
-    """
-    RpcClient represents an WS RPC client, tailored to the
-    QIX Associative Engine's RPC API specifically.
-
-    It is safe to use the client.send from multiple threads.
-    """
+    _socket = None
+    _watch_recv_thread = None
 
     def __init__(
-        self,
-        url: str,
-        header: typing.List[str] = None,
-        request_interceptors: typing.List[RequestInterceptor] = None,
-        response_interceptors: typing.List[ResponseInterceptor] = None,
+        self, ws_url: str, headers: List[str] = None, interceptors: Interceptors = None
     ):
-        self._socket = None
-        self._id = -1
-        self._received = None
-        self._recv_error = None
-
-        if header is None:
-            header = []
-        self._response_interceptors = [RetryResponseInterceptor(self).intercept]
-        if response_interceptors:
-            self._response_interceptors.extend(response_interceptors)
-        self._request_interceptors = []
-        if request_interceptors:
-            self._request_interceptors.extend(request_interceptors)
-        self.lock = Lock()
-        if not url:
+        if headers is None:
+            headers = []
+        if not ws_url:
             raise Exception("Empty url")
-        self.connect(url, header)
+        self._headers = headers
+        self._ws_url = ws_url
+        self._interceptors = interceptors
+        self.lock = Lock()
 
     def _watch_recv(self):
         """
@@ -89,27 +126,19 @@ class RpcClient:
                     # wait_response will never finish
                     self._received_added.notify_all()
 
-    def connect(self, url: str, headers: typing.List[str] = None):
+    def open(self):
         """
         connect establishes a connection to provided url
         using the specified headers.
 
         If the client is already connected an exception will
         be raised.
-
-        Parameters
-        ----------
-        url: string the websocket url
-        headers: list[str] headers
         """
-        if headers is None:
-            headers = []
-
         if self.is_connected():
             raise Exception("Client already connected")
         socket = websocket.WebSocket()
         try:
-            socket.connect(url, header=headers, suppress_origin=True)
+            socket.connect(self._ws_url, header=self._headers, suppress_origin=True)
         except Exception as exc:
             raise FailedToConnect() from exc
 
@@ -120,12 +149,12 @@ class RpcClient:
 
         self._watch_recv_thread = Thread(target=self._watch_recv)
         self._watch_recv_thread.start()
+        return self
 
     def is_connected(self):
         """
         return connected state
         """
-
         return self._socket and self._socket.connected
 
     def close(self):
@@ -135,7 +164,7 @@ class RpcClient:
 
         if self.is_connected():
             self._socket.send_close()
-        if self._watch_recv_thread.is_alive():
+        if self._watch_recv_thread is not None and self._watch_recv_thread.is_alive():
             self._watch_recv_thread.join()
 
     def __enter__(self):
@@ -179,20 +208,17 @@ class RpcClient:
         for param in params:
             encoded_params.append(param)
 
-        data = {
-            "jsonrpc": "2.0",
-            "id": self._id,
-            "method": method,
-            "handle": handle,
-            "params": encoded_params,
-        }
+        data = RequestObject(
+            id=self._id, method=method, handle=handle, params=encoded_params
+        )
+
         # send and wait respons
-        data = reduce(lambda d, f: f(d), self._request_interceptors, data)
-        json_data = json.dumps(data)
+        data = reduce(lambda d, f: f(d), self._interceptors["request"].handlers, data)
+        json_data = json.dumps(asdict(data))
         self._socket.send(json_data)
         res = self._wait_response(id_)
         res["request_data"] = data
-        res = reduce(lambda r, f: f(r), self._response_interceptors, res)
+        res = reduce(lambda r, f: f(r), self._interceptors["response"].handlers, res)
         return_value = None
         if "result" in res:
             return_value = res["result"]
@@ -219,3 +245,38 @@ class RpcClient:
             res = self._received[id_]
             del self._received[id_]
             return res
+
+
+class RpcClient:
+    __config: Config
+    # property for storing the interceptors
+    interceptors: Interceptors
+
+    sessions: Dict[str, RpcSession] = {}
+
+    def __init__(self, config) -> None:
+        self.__config = config
+        # initiating the interceptors
+        self.interceptors = dict(
+            request=InterceptorHandler[RequestInterceptor](),
+            response=InterceptorHandler[ResponseInterceptor](),
+        )
+
+    def rpc(self, app_id: str) -> RpcSession:
+        hostname = urlparse(self.__config.host).hostname
+        ws_url = "wss://" + hostname.strip("/") + "/app/" + app_id
+        headers = ["Authorization: Bearer %s" % self.__config.api_key]
+        if ws_url not in self.sessions:
+            self.sessions[ws_url] = RpcSession(ws_url, headers, self.interceptors)
+        return self.sessions[ws_url]
+
+
+class RpcClientInstance:
+    interceptors: Interceptors
+
+    def __init__(self, rpcClient: RpcClient) -> None:
+        self._rpcClient = rpcClient
+        self.interceptors = rpcClient.interceptors
+
+    def __call__(self, app_id: str) -> RpcSession:
+        return self._rpcClient.rpc(app_id)
